const Ast = @import("./kym/Ast.zig");

const Chunk = @import("./kym/Chunk.zig");

const Table = @import("./kym/Table.zig");

const coral = @import("coral");

const file = @import("./file.zig");

pub const Any = union (enum) {
	nil,
	boolean: bool,
	number: Float,
	string: []const coral.io.Byte,
	dynamic: *DynamicObject,

	pub fn expect_dynamic(self: Any) ?*DynamicObject {
		return switch (self) {
			.dynamic => |dynamic| dynamic,
			else => null,
		};
	}

	pub fn expect_number(self: Any) ?Float {
		return switch (self) {
			.number => |number| number,
			else => null,
		};
	}

	pub fn expect_string(self: Any) ?[]const coral.io.Byte {
		return switch (self) {
			.string => |string| string,
			else => null,
		};
	}
};

pub const Caller = coral.io.Generator(RuntimeError!?*RuntimeRef, *RuntimeEnv);

pub const DynamicObject = struct {
	userdata: []coral.io.Byte,
	typeinfo: *const Typeinfo,

	pub fn as_caller(self: *DynamicObject) Caller {
		return Caller.bind(DynamicObject, self, DynamicObject.call);
	}

	fn call(self: *DynamicObject, env: *RuntimeEnv) RuntimeError!?*RuntimeRef {
		return self.typeinfo.call(env);
	}
};

pub const ErrorHandler = coral.io.Generator(void, ErrorInfo);

pub const ErrorInfo = struct {
	message: []const coral.io.Byte,
	frames: []const Frame,
};

pub const Float = f64;

pub const Frame = struct {
	name: []const coral.io.Byte,
	arg_count: u8,
	locals_top: usize,
};

pub const IndexContext = struct {
	env: *RuntimeEnv,
	userdata: []coral.io.Byte,
	index_ref: ?*const RuntimeRef,
};

pub const RuntimeEnv = struct {
	allocator: coral.io.Allocator,
	error_handler: ErrorHandler,
	syscallers: SyscallerTable,
	local_refs: RefStack,
	frames: FrameStack,
	ref_values: RefSlab,

	const FrameStack = coral.list.Stack(Frame);

	const SyscallerTable = coral.map.StringTable(Caller);

	const RefStack = coral.list.Stack(?*RuntimeRef);

	const RefSlab = coral.map.Slab(struct {
		ref_count: usize,

		object: union (enum) {
			false,
			true,
			number: Float,
			string: []coral.io.Byte,
			dynamic: *DynamicObject,
		},
	});

	pub const Syscall = struct {
		name: []const coral.io.Byte,
		caller: Caller,
	};

	pub fn acquire(self: *RuntimeEnv, ref: ?*const RuntimeRef) ?*RuntimeRef {
		const key = @intFromPtr(ref orelse return null);
		var ref_data = self.ref_values.remove(key);

		coral.debug.assert(ref_data != null);

		ref_data.?.ref_count += 1;

		coral.debug.assert(self.ref_values.insert_at(key, ref_data.?));

		return @ptrFromInt(key);
	}

	pub fn bind_syscalls(self: *RuntimeEnv, syscalls: []const Syscall) RuntimeError!void {
		for (syscalls) |syscall| {
			_ = try self.syscallers.replace(syscall.name, syscall.caller);
		}
	}

	pub fn call(
		self: *RuntimeEnv,
		name: []const coral.io.Byte,
		arg_count: u8,
		caller: Caller,
	) RuntimeError!?*RuntimeRef {
		try self.frames.push_one(.{
			.name = name,
			.arg_count = arg_count,
			.locals_top = self.local_refs.values.len,
		});

		defer {
			const frame = self.frames.pop();

			coral.debug.assert(frame != null);

			{
				var pops_remaining = (self.local_refs.values.len - frame.?.locals_top) + frame.?.arg_count;

				while (pops_remaining != 0) : (pops_remaining -= 1) {
					const local = self.local_refs.pop();

					coral.debug.assert(local != null);

					self.discard(local.?);
				}
			}
		}

		return caller.invoke(self);
	}

	pub fn discard(self: *RuntimeEnv, ref: ?*RuntimeRef) void {
		const key = @intFromPtr(ref orelse return);
		var ref_data = self.ref_values.remove(key) orelse unreachable;

		coral.debug.assert(ref_data.ref_count != 0);

		ref_data.ref_count -= 1;

		if (ref_data.ref_count == 0) {
			switch (ref_data.object) {
				.string => |string| self.allocator.deallocate(string),

				.dynamic => |dynamic| {
					dynamic.typeinfo.clean(self, dynamic.userdata);
					self.allocator.deallocate(dynamic.userdata);
					self.allocator.deallocate(dynamic);
				},

				else => {},
			}
		} else {
			coral.debug.assert(self.ref_values.insert_at(key, ref_data));
		}
	}

	pub fn execute_file(self: *RuntimeEnv, file_access: file.Access, file_path: file.Path) RuntimeError!?*RuntimeRef {
		if ((try file.allocate_and_load(self.allocator, file_access, file_path))) |file_data| {
			defer self.allocator.deallocate(file_data);

			if (file_path.to_string()) |string| {
				return self.execute_script(string, file_data);
			}
		}

		return self.raise(error.BadOperation, "failed to load file");
	}

	pub fn execute_script(
		self: *RuntimeEnv,
		name: []const coral.io.Byte,
		data: []const coral.io.Byte,
	) RuntimeError!?*RuntimeRef {
		var ast = Ast.make(self.allocator, name);

		defer ast.free();

		ast.parse(data) catch |parse_error| switch (parse_error) {
			error.BadSyntax => return self.raise(error.BadSyntax, ast.error_message()),
			error.OutOfMemory => return error.OutOfMemory,
		};

		var chunk = Chunk.make(self);

		defer chunk.free();

		try chunk.compile_ast(ast);

		return self.call(name, 0, chunk.as_caller());
	}

	pub fn free(self: *RuntimeEnv) void {
		for (self.local_refs.values) |ref| {
			self.discard(ref);
		}

		self.frames.free();
		self.syscallers.free();
		self.local_refs.free();
		self.ref_values.free();
	}

	pub fn get_local(self: *RuntimeEnv, local: u8) RuntimeError!?*RuntimeRef {
		if (local >= self.local_refs.values.len) {
			return self.raise(error.IllegalState, "out of bounds local get");
		}

		return self.acquire(self.local_refs.values[local]);
	}

	pub fn pop_local(self: *RuntimeEnv) RuntimeError!?*RuntimeRef {
		return self.local_refs.pop() orelse self.raise(error.IllegalState, "stack underflow");
	}

	pub fn push_boolean(self: *RuntimeEnv, boolean: bool) RuntimeError!void {
		return self.local_refs.push_one(try self.new_boolean(boolean));
	}

	pub fn push_number(self: *RuntimeEnv, number: Float) RuntimeError!void {
		return self.local_refs.push_one(try self.new_number(number));
	}

	pub fn push_ref(self: *RuntimeEnv, ref: ?*RuntimeRef) RuntimeError!void {
		return self.local_refs.push_one(self.acquire(ref));
	}

	pub fn set_local(self: *RuntimeEnv, local: u8, value: ?*RuntimeRef) RuntimeError!void {
		if (local >= self.local_refs.values.len) {
			return self.raise(error.IllegalState, "out of bounds local set");
		}

		self.local_refs.values[local] = self.acquire(value);
	}

	pub fn make(allocator: coral.io.Allocator, error_handler: ErrorHandler) coral.io.AllocationError!RuntimeEnv {
		return RuntimeEnv{
			.local_refs = RefStack.make(allocator),
			.ref_values = RefSlab.make(allocator),
			.frames = FrameStack.make(allocator),
			.syscallers = SyscallerTable.make(allocator),
			.error_handler = error_handler,
			.allocator = allocator,
		};
	}

	pub fn new_boolean(self: *RuntimeEnv, value: bool) RuntimeError!?*RuntimeRef {
		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = if (value) .true else .false,
		}));
	}

	pub fn new_dynamic(
		self: *RuntimeEnv,
		userdata: []const coral.io.Byte,
		typeinfo: *const Typeinfo,
	) RuntimeError!*RuntimeRef {
		const userdata_copy = try coral.io.allocate_copy(self.allocator, userdata);

		errdefer self.allocator.deallocate(userdata_copy);

		const dynamic_object = try coral.io.allocate_one(self.allocator, DynamicObject{
			.typeinfo = typeinfo,
			.userdata = userdata_copy,
		});

		errdefer self.allocator.deallocate(dynamic_object);

		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.dynamic = dynamic_object},
		}));
	}

	pub fn new_number(self: *RuntimeEnv, value: Float) RuntimeError!?*RuntimeRef {
		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.number = value},
		}));
	}

	pub fn new_string(self: *RuntimeEnv, string_data: []const coral.io.Byte) RuntimeError!?*RuntimeRef {
		const string_copy = try coral.io.allocate_copy(self.allocator, string_data);

		errdefer self.allocator.deallocate(string_copy);

		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.string = string_copy},
		}));
	}

	pub fn raise(self: *RuntimeEnv, error_value: RuntimeError, message: []const coral.io.Byte) RuntimeError {
		self.error_handler.invoke(.{
			.message = message,
			.frames = self.frames.values,
		});

		return error_value;
	}

	pub fn syscaller(self: *RuntimeEnv, name: []const coral.io.Byte) RuntimeError!Caller {
		return self.syscallers.lookup(name) orelse self.raise(error.BadOperation, "attempt to call undefined syscall");
	}

	pub fn unbox(self: *RuntimeEnv, ref: ?*const RuntimeRef) Any {
		if (ref) |live_ref| {
			const ref_data = self.ref_values.lookup(@intFromPtr(live_ref));

			coral.debug.assert(ref_data != null);

			return switch (ref_data.?.object) {
				.false => .{.boolean = false},
				.true => .{.boolean = true},
				.number => |number| .{.number = number},
				.string => |string| .{.string = string},
				.dynamic => |dynamic| .{.dynamic = dynamic},
			};
		}

		return .nil;
	}

	pub fn view_arg(self: *RuntimeEnv, index: usize) RuntimeError!?*const RuntimeRef {
		const frame = self.frames.peek() orelse return self.raise(error.IllegalState, "stack underflow");

		if (index >= frame.arg_count) {
			return null;
		}

		return self.local_refs.values[frame.locals_top - (1 + index)];
	}
};

pub const RuntimeError = coral.io.AllocationError || error {
	IllegalState,
	TypeMismatch,
	BadOperation,
	BadSyntax,
};

pub const RuntimeRef = opaque {};

pub const TestContext = struct {
	env: *RuntimeEnv,
	userdata: []const coral.io.Byte,
	other_ref: ?*const RuntimeRef,
};

pub const Typeinfo = struct {
	name: []const coral.io.Byte,
	call: *const fn (env: *RuntimeEnv) RuntimeError!?*RuntimeRef = default_call,
	clean: *const fn (env: *RuntimeEnv, userdata: []coral.io.Byte) void = default_clean,
	get: *const fn (context: IndexContext) RuntimeError!?*RuntimeRef = default_get,
	set: *const fn (context: IndexContext, value: ?*const RuntimeRef) RuntimeError!void = default_set,
	test_difference: *const fn (context: TestContext) RuntimeError!Float = default_test_difference,
	test_equality: *const fn (context: TestContext) RuntimeError!bool = default_test_equality,

	fn default_call(env: *RuntimeEnv) RuntimeError!?*RuntimeRef {
		return env.raise(error.TypeMismatch, "object is not callable");
	}

	fn default_clean(_: *RuntimeEnv, _: []coral.io.Byte) void {
		// Nothing to clean by default.
	}

	fn default_get(context: IndexContext) RuntimeError!?*RuntimeRef {
		return context.env.raise(error.TypeMismatch, "object is not indexable");
	}

	fn default_set(context: IndexContext, _: ?*const RuntimeRef) RuntimeError!void {
		return context.env.raise(error.TypeMismatch, "object is not indexable");
	}

	fn default_test_difference(context: TestContext) RuntimeError!Float {
		return context.env.raise(error.TypeMismatch, "object is not comparable");
	}

	fn default_test_equality(context: TestContext) RuntimeError!bool {
		return switch (context.env.unbox(context.other_ref)) {
			.dynamic => |dynamic| context.userdata.ptr == dynamic.userdata.ptr,
			else => false,
		};
	}
};

pub fn call(
	env: *RuntimeEnv,
	caller_ref: ?*const RuntimeRef,
	callable_ref: ?*const RuntimeRef,
	arg_refs: []const ?*const RuntimeRef,
) RuntimeError!?*RuntimeRef {
	for (arg_refs) |arg_ref| {
		try env.local_push_ref(arg_ref);
	}

	env.frame_push("", arg_refs.len);

	defer env.frame_pop();

	const dynamic = try env.unbox(callable_ref).expect_dynamic();

	return dynamic.type_info.call(.{
		.env = env,
		.caller = caller_ref,
		.userdata = dynamic.userdata,
	});
}

pub fn get(
	env: *RuntimeEnv,
	indexable_ref: ?*const RuntimeRef,
	index_ref: ?*const RuntimeRef,
) RuntimeError!?*RuntimeRef {
	const dynamic = try unbox_dynamic(env, indexable_ref);

	return dynamic.typeinfo.get(.{
		.userdata = dynamic.userdata,
		.env = env,
		.index_ref = index_ref,
	});
}

pub fn get_field(
	env: *RuntimeEnv,
	indexable_ref: ?*const RuntimeRef,
	field_name: []const coral.io.Byte,
) RuntimeError!?*RuntimeRef {
	const field_name_ref = try env.new_string(field_name);

	defer env.discard(field_name_ref);

	return get(env, indexable_ref, field_name_ref);
}

pub fn set(
	env: *RuntimeEnv,
	indexable_ref: ?*const RuntimeRef,
	index_ref: ?*const RuntimeRef,
	value_ref: ?*const RuntimeRef,
) RuntimeError!void {
	const dynamic = try unbox_dynamic(env, indexable_ref);

	return dynamic.typeinfo.set(.{
		.userdata = dynamic.userdata,
		.env = env,
		.index_ref = index_ref,
	}, value_ref);
}

pub fn set_field(
	env: *RuntimeEnv,
	indexable_ref: ?*const RuntimeRef,
	field_name: []const coral.io.Byte,
	value_ref: ?*const RuntimeRef,
) RuntimeError!void {
	const field_name_ref = try env.new_string(field_name);

	defer env.discard(field_name_ref);

	return set(env, indexable_ref, field_name_ref, value_ref);
}

pub const new_table = Table.new;

pub fn test_difference(env: *RuntimeEnv, lhs_ref: ?*const RuntimeRef, rhs_ref: ?*const RuntimeRef) RuntimeError!Float {
	return switch (env.unbox(lhs_ref)) {
		.nil => env.raise(error.TypeMismatch, "cannot compare nil objects"),
		.boolean => env.raise(error.TypeMismatch, "cannot compare boolean objects"),

		.number => |lhs_number| switch (env.unbox(rhs_ref)) {
			.number => |rhs_number| rhs_number - lhs_number,
			else => env.raise(error.TypeMismatch, "right-hand object is not comparable with number objects"),
		},

		.string => |lhs_string| switch (env.unbox(rhs_ref)) {
			.string => |rhs_string| @floatFromInt(coral.io.compare(lhs_string, rhs_string)),
			else => env.raise(error.TypeMismatch, "right-hand object is not comparable with string objects"),
		},

		.dynamic => |lhs_dynamic| lhs_dynamic.typeinfo.test_difference(.{
			.env = env,
			.userdata = lhs_dynamic.userdata,
			.other_ref = rhs_ref,
		}),
	};
}

pub fn test_equality(env: *RuntimeEnv, lhs_ref: ?*const RuntimeRef, rhs_ref: ?*const RuntimeRef) RuntimeError!bool {
	return switch (env.unbox(lhs_ref)) {
		.nil => (rhs_ref == null),

		.boolean => |lhs_boolean| switch (env.unbox(rhs_ref)) {
			.boolean => |rhs_boolean| rhs_boolean == lhs_boolean,
			else => false,
		},

		.number => |lhs_number| switch (env.unbox(rhs_ref)) {
			.number => |rhs_number| lhs_number == rhs_number,
			else => false,
		},

		.string => |lhs_string| switch (env.unbox(rhs_ref)) {
			.string => |rhs_string| coral.io.equals(lhs_string, rhs_string),
			else => false,
		},

		.dynamic => |lhs_dynamic| try lhs_dynamic.typeinfo.test_equality(.{
			.env = env,
			.userdata = lhs_dynamic.userdata,
			.other_ref = rhs_ref,
		}),
	};
}

pub fn unbox_dynamic(env: *RuntimeEnv, ref: ?*const RuntimeRef) RuntimeError!*DynamicObject {
	return env.unbox(ref).expect_dynamic() orelse env.raise(error.TypeMismatch, "expected dynamic object");
}

pub fn unbox_string(env: *RuntimeEnv, ref: ?*const RuntimeRef) RuntimeError![]const coral.io.Byte {
	return env.unbox(ref).expect_string() orelse env.raise(error.TypeMismatch, "expected string object");
}
